function usage() {
    echo ""
    echo "########### Build Script for the analysis code #########"
    echo
    echo "Author: Lenka Pitonakova - contact@lenkaspace.net"
    echo " "
    echo "Usage: ./build.sh -c"
    echo "  Compiles and globally installs the CPP code using pip"
    echo " "
    echo "Additionally, you can use these options: "
    echo "  -r path: Run a python script"
    echo " "
    echo "Examples:"
    echo "  ./build.sh -c"
    echo "  ./build.sh -r python/s1_corridor.py"
    echo
    echo "###########################################"
    echo ""
}


function compile() {
    echo ""
    echo ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Compiling ..."
	
	#-- set variables that the python setup script can read
	export PROJECT_DIRECTORY_PATH="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && cd ../ && pwd )"
    
    #-- compile and build
    pip install ./cpp --user

}


function run() {
	echo ""
    echo ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Running ${RUN_PATH}..."
    
    python ${RUN_PATH}
    
}

########################################################################


RUN=0
HELP=1
COMPILE=0
RUN_PATH=""

#-- check if help should be displayed
while getopts "hr:c" name
do
    case $name in
        h)
        HELP=1;;
        c)
        COMPILE=1
        HELP=0;;
        r)
        RUN=1
        HELP=0;
        RUN_PATH="$OPTARG";;
        

        \?) echo "!!! Invalid option: -"$OPTARG"" >&2
        exit 1;;
        : ) echo "!!! Option -"$OPTARG" requires an argument." >&2
        exit 1;;

    esac
done

if [ $HELP -eq 1 ]; then
    usage;
    exit 1;
fi

#-- compile
if [ $COMPILE -eq 1 ]; then
    compile
fi


#-- exit if not running
if [ $RUN -eq 0 ]; then
    exit 1;
fi

run
