/* ================================================================
 * ================================================================
 *
 * Input / output static helper functions
 *
 * Author: L. Pitonakova (http://lenkaspace.net)
 * License: GNU General Public License. Please credit me when using my work.
 *
 * ================================================================
 * ================================================================
 */

#include "ioHelper.h"

#include <iterator>
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept>

IOHelper::IOHelper() {


}

IOHelper::~IOHelper() {

}


/**
 * Read a file that consists of lines and delimiter-separated float values and load it into a 2D array
 * where 0th dimension is rows, 1st dimension is columns
 */
void IOHelper::FileToVector(std::string filePath_, std::vector<std::vector<float>>& resultArray_, bool ignoreFirstLine_, const char delimiter_, float defaultValue_) {
   //printf("Opening file %s (ignoring first line: %d)\n",filePath_.c_str(), ignoreFirstLine_);

   std::ifstream fileStream (filePath_);
   std::string line;

   bool didPassFirstLine = false;
   bool hasLine = false;
   while (std::getline(fileStream, line)) {
      if (!didPassFirstLine && ignoreFirstLine_) {
         didPassFirstLine = true;
         continue;
      }
      std::vector<float> rowData;

      std::stringstream lineStream(line);
      std::string cell;
      try {
         while(std::getline(lineStream, cell, delimiter_)) {
            hasLine = true;
            rowData.push_back(std::stof(cell));
         }

         //-- add empty element if there was a trailing separator
         if (!lineStream && cell.empty()) {
            rowData.push_back(defaultValue_);
         }
         resultArray_.push_back(rowData);
      } catch (...) {

      }
   }

   //-- check that the file was open and non-empty
   if (!hasLine) {
      std::stringstream text;
      text << "File " << filePath_ << " does not exist or is empty";
      throw std::invalid_argument(text.str());
   }
}

/* ---- overloading variants ------ */
void IOHelper::FileToVector(std::string filePath_, std::vector<std::vector<float>>& resultArray_) {
   IOHelper::FileToVector(filePath_, resultArray_, '\t');
}

void IOHelper::FileToVector(std::string filePath_, std::vector<std::vector<float>>& resultArray_, bool ignoreFirstLine_) {
   IOHelper::FileToVector(filePath_, resultArray_, ignoreFirstLine_, '\t', -1);
}


void IOHelper::FileToVector(std::string filePath_, std::vector<std::vector<float>>& resultArray_, const char delimiter_) {
   IOHelper::FileToVector(filePath_, resultArray_, false, '\t', -1);
}
/* ---- ------------------- ------ */





/**
 * Read a file that consists of lines and delimiter-separated float values and load it into a 2D array
 * where 0th dimension is rows, 1st dimension is columns
 */
void IOHelper::FileToVector(std::string filePath_, std::vector<std::vector<std::string>>& resultArray_, bool ignoreFirstLine_, const char delimiter_, std::string defaultValue_) {
   //printf("Opening file %s (ignoring first line: %d)\n",filePath_.c_str(), ignoreFirstLine_);

   std::ifstream fileStream (filePath_);
   std::string line;

   bool didPassFirstLine = false;
   bool hasLine = false;
   while (std::getline(fileStream, line)) {
      if (!didPassFirstLine && ignoreFirstLine_) {
         didPassFirstLine = true;
         continue;
      }
      std::vector<std::string> rowData;

      std::stringstream lineStream(line);
      std::string cell;
      try {
         while(std::getline(lineStream, cell, delimiter_)) {
            hasLine = true;
            rowData.push_back(cell);
         }

         //-- add empty element if there was a trailing separator
         if (!lineStream && cell.empty()) {
            rowData.push_back(defaultValue_);
         }
         resultArray_.push_back(rowData);
      } catch (...) {

      }
   }

   //-- check that the file was open and non-empty
   if (!hasLine) {
      std::stringstream text;
      text << "File " << filePath_ << " does not exist or is empty";
      throw std::invalid_argument(text.str());
   }
}

/* ---- overloading variants ------ */
void IOHelper::FileToVector(std::string filePath_, std::vector<std::vector<std::string>>& resultArray_) {
   IOHelper::FileToVector(filePath_, resultArray_, '\t');
}

void IOHelper::FileToVector(std::string filePath_, std::vector<std::vector<std::string>>& resultArray_, bool ignoreFirstLine_) {
   IOHelper::FileToVector(filePath_, resultArray_, ignoreFirstLine_, '\t', "");
}

void IOHelper::FileToVector(std::string filePath_, std::vector<std::vector<std::string>>& resultArray_, const char delimiter_) {
   IOHelper::FileToVector(filePath_, resultArray_, false, '\t', "");
}
/* ---- ------------------- ------ */
