####################################################################
##
## Various helper functions that other files use
##
## Author: L. Pitonakova (http://lenkaspace.net)
## License: GNU General Public License. Please credit me when using my work.
##
###################################################################

# import math;
import numpy;
import os;


import gaData_module;

import constants;
import visualization;

CACHE_DEBUG_OUTPUT = False;

def createFolder(folderPath_):
    #folderPath_ = constants.BASE_FILE_PATH_OUTPUT_IMAGES + "/" + folderPath_
    if not os.path.exists(folderPath_):
        print("CREATED FOLDER " + folderPath_)
        os.makedirs(folderPath_)



def generateXTickPositionsAndLabels_time(data_, startTime_, endTime_=-1, secondsPerTick_ = 10, nonRegularTimes_=[], nonRegularTimesStringFormat_="{}"):
    xTickLabelPositions = numpy.arange(0,len(data_), secondsPerTick_*10); #each minute is step of 600 (ticks, 10 ticks = 1 second)

    if (endTime_ > 0):
        xTickLabelPositions = numpy.append(xTickLabelPositions,[endTime_])

    if (len(nonRegularTimes_) > 0):
        xTickLabelPositions = numpy.sort(numpy.append(xTickLabelPositions, nonRegularTimes_));


    xTickLabels = []
    for t in range(len(xTickLabelPositions)):
        timeInSeconds = (startTime_ + xTickLabelPositions[t])/10;
        if (xTickLabelPositions[t] in nonRegularTimes_):
            xTickLabels.append(nonRegularTimesStringFormat_.format(timeInSeconds));
        else:
            xTickLabels.append(str(timeInSeconds));
    
    return(xTickLabelPositions, xTickLabels);





def generateXTickPositionsAndLabels(data_, startLabelValue_, endLabelValue_, numOfElemets_ = 10, nonRegularPositions_=[], nonRegularPositionsStringFormat_="{}"):
    xTickLabelPositions = numpy.arange(0,len(data_), len(data_)/numOfElemets_);
    if (len(nonRegularPositions_) > 0):
        xTickLabelPositions = numpy.sort(numpy.append(xTickLabelPositions, nonRegularPositions_));

    xTickLabels = [];
    xTickLabels = numpy.arange(startLabelValue_, endLabelValue_, (endLabelValue_ - startLabelValue_)  /numOfElemets_);
    xTickLabels = [ round(xTickLabels[x],1) for x in range(len(xTickLabels))];
    #for t in range(len(xTickLabelPositions)):
    #    labelVal = round(startLabelValue_ + (endLabelValue_ - startLabelValue_)* (t/len(xTickLabelPositions)),2);
    #    print(" bin no " + str(t) + "  val " + str(labelVal))
    #    if (xTickLabelPositions[t] in nonRegularPositions_):
    #        xTickLabels.append(nonRegularPositionsStringFormat_.format(labelVal));
    #    else:
    #        xTickLabels.append(str(t))
    return(xTickLabelPositions, xTickLabels);







def calculatePassLength(config_ : visualization.PlotConfig):
    """
    :return: (average, maximum) pass length in seconds in a set of results
    """


    if ("s1_x1_room" in config_.outputDirectoryPath): # code for s1_x1 experiment did not yet have reporting of robots reaching end loop waypoint
        return (500, 500)


    averagePassLength = 0;
    totalNumOfPasses = 0;
    maxPassLength = 0;
    averageEnvChangedTime = 0;

    envChangedTimesData = gaData_module.getEventTimes(config_.dataDirectoryPath, config_.numOfRuns, config_.numOfRobots, 1);
    if ("s1_x1_corridor" in config_.outputDirectoryPath):
        # code for s1_x1 experiment did not yet have reporting of robots reaching end loop waypoint
        allRobotsReachedEndLoopData = gaData_module.getRobotTeleportTimes(config_.dataDirectoryPath, config_.numOfRuns, config_.numOfRobots);
    else:
        # in all other experiments, times at which robot passes the end loop waypoint are noted down.
        allRobotsReachedEndLoopData = gaData_module.getRobotReachedWaypointTimes(config_.dataDirectoryPath, config_.numOfRuns, config_.numOfRobots, constants.WAYPOINT_ID_END_LOOP);

    for robotId in range(config_.numOfRobots):
        for runNo in range(config_.numOfRuns):
            totalNumOfPasses += len(allRobotsReachedEndLoopData[robotId][runNo]);
            for i in range(len(allRobotsReachedEndLoopData[robotId][runNo])):
                if (i == 0):
                    passLength = allRobotsReachedEndLoopData[robotId][runNo][i];
                else:
                    passLength = allRobotsReachedEndLoopData[robotId][runNo][i] - allRobotsReachedEndLoopData[robotId][runNo][i - 1];
                averagePassLength += passLength;
                maxPassLength = (int)(max(maxPassLength, passLength));

            #print(runNo, len(envChangedTimesData), len(envChangedTimesData[runNo]))
            averageEnvChangedTime += (int)(envChangedTimesData[runNo][0]);

    averagePassLength /= totalNumOfPasses;
    averageEnvChangedTime /= config_.numOfRuns;
    #print("MAX PASS LENGTH", maxPassLength);
    #print("AVG ENVIRONMENT CHANGE TIME", averageEnvChangedTime);

    return(averagePassLength, maxPassLength);


def calculateTotalExperimentTime(config_ : visualization.PlotConfig):
    """
    :param config_:
    :return: Total experiment time in seconds
    """

    #-- try to extract numeric value that follows the keyword for robot maximum speed
    if ("bcn" in config_.dataDirectoryPath):
        return 4000;

    elif ("rms" in config_.dataDirectoryPath):
        valueStartIndex = config_.dataDirectoryPath.find("rms") + 3
        valueEndIndex = config_.dataDirectoryPath.find("_", valueStartIndex);
        if (valueEndIndex == -1):
            valueEndIndex = len(config_.dataDirectoryPath);
        rmsVal = (int)(config_.dataDirectoryPath[(int)(valueStartIndex): (int)(valueEndIndex)]);

        #-- following calculation of total experiment time from the run .sh scripts, calculate the total experiment time
        return (int)((40*10*1.0)/(rmsVal/100.0));

    return config_.totalExperimentTime;




def getHumanReadableParameterString(paramString_):
    """

    :param paramString_:
    :return: Parameter string with parameter shortcuts being replaced by mathematical notation
    """

    returnString = paramString_.replace("_",", ");
    returnString = returnString.replace("inw", '$N_W$=')
    returnString = returnString.replace("at", '$\Theta_A$=')

    return returnString;

