####################################################################
##
## Code for comparing selected parameter sets.
##
## Author: L. Pitonakova (http://lenkaspace.net)
## License: GNU General Public License. Please credit me when using my work.
##
####################################################################


#--------------------------------
#      NETWORK METRICS - comparing different sets
#--------------------------------

import gaData_module;
from string import ascii_lowercase
from pyCreeper import crGraphs;
from pyCreeper import crGraphStyle;
from pyCreeper import crData;
import numpy;
import helpers;
import constants;
import visualization;
import paths;



config = visualization.PlotConfig()

ANALYSIS_NAME = "corridor";
#ANALYSIS_NAME = "room";


if ("corridor" in ANALYSIS_NAME):
   OPTIMIZED_INW_AT_SMALL = "inw12_at0.8";
   #OPTIMIZED_INW_AT_LARGE = "inw36_at0.2";
   FOUND_INW_AT = "inw39_at0.2";
   #OPTIMIZED_SET_BASIC = OPTIMIZED_INW_AT + "_ht0.30"
else:
   OPTIMIZED_INW_AT_SMALL = "inw9_at0.7";
   #OPTIMIZED_INW_AT_LARGE = "inw24_at0.3";
   FOUND_INW_AT = "inw30_at0.2";
   #OPTIMIZED_SET_BASIC = OPTIMIZED_INW_AT + "_ht0.30"


#DO_DISHABITUATION = False;
DO_ADAPTIVE_INW = False;
DO_BLOB_CAMERA_NOIZE = False;
DO_WINNING_NETWORKS = True;


DO_NOVELTY_REPORTING_OVER_PASSES = False;
DO_NEURON_ERRORS = True;

config.numOfRuns = 50;

#-- define parameter sets
#if (DO_DISHABITUATION):
#   PARAM_SETS = [OPTIMIZED_SET_BASIC, OPTIMIZED_INW_AT + "_dishab_unnp0.002", OPTIMIZED_INW_AT + "_dishab_unnp0.003"];
#   COMPARISON_NAME = "dishabituation";

if (DO_ADAPTIVE_INW):
   #PARAM_SETS = [OPTIMIZED_SET_BASIC, "inwAdaptive_at0.2_ht0.30", "inw66_at0.2"];
   PARAM_SETS = ["s1_x1_" + ANALYSIS_NAME + "/" + OPTIMIZED_INW_AT_SMALL, "s1_x1_" + ANALYSIS_NAME + "/" + FOUND_INW_AT, "s1_x1_" + ANALYSIS_NAME + "/inwAdaptive_at0.2",
               "s1_x2_" + ANALYSIS_NAME + "/" + OPTIMIZED_INW_AT_SMALL + "_rms20", "s1_x2_" + ANALYSIS_NAME + "/" + FOUND_INW_AT + "_rms20", "s1_x2_" + ANALYSIS_NAME + "/inwAdaptive_at0.2_rms20"];
   COMPARISON_NAME = "adaptiveInw";

elif (DO_BLOB_CAMERA_NOIZE):
   PARAM_SETS = ['s1_x2_' + ANALYSIS_NAME + "/" + OPTIMIZED_INW_AT_SMALL + "_rms20",'s1_x2_' + ANALYSIS_NAME + "/" + FOUND_INW_AT + "_rms20",'s1_x2_' + ANALYSIS_NAME + "/inwAdaptive_at0.2_rms20",
               's1_x2_' + ANALYSIS_NAME + "/" + OPTIMIZED_INW_AT_SMALL + "_rms20_bcn0.150",'s1_x2_' + ANALYSIS_NAME + "/" + FOUND_INW_AT + "_rms20_bcn0.150",'s1_x2_' + ANALYSIS_NAME + "/inwAdaptive_at0.2_rms20_bcn0.150"];
   COMPARISON_NAME = "blobCameraNoize";

elif (DO_WINNING_NETWORKS):
   PARAM_SETS = ["s1_x2_" + ANALYSIS_NAME + "/" + OPTIMIZED_INW_AT_SMALL + "_rms20", "s1_x2_" + ANALYSIS_NAME + "/" + FOUND_INW_AT + "_rms20", "s1_x2_" + ANALYSIS_NAME + "/inwAdaptive_at0.2_rms20"];
   COMPARISON_NAME = "winningNetworks";


#-- define and create directories
#DATA_DIRECTORY = paths.BASE_FILE_PATH_DATA + "/" + ANALYSIS_NAME;
DATA_DIRECTORY = paths.BASE_FILE_PATH_DATA;
OUTPUT_DIRECTORY = paths.BASE_FILE_PATH_OUTPUT_IMAGES + "/" + ANALYSIS_NAME + "__setComparison";
helpers.createFolder(OUTPUT_DIRECTORY)


#-- define legend labels
LETTERS = ascii_lowercase;
config.style.markers = ['cs','bs','rs', 'cd','bd','rd'];
config.style.lineStyles = ['-','-','--','-','-','--'];
config.outputDirectoryPath = OUTPUT_DIRECTORY + "/" + COMPARISON_NAME

legendLabels = ["(" + LETTERS[l] + ")" for l in range(len(PARAM_SETS))];


#-- get data
numOfNoveltyTruePositivesData = [] ;  # each data point is collated number of true positives in a certain pass and run, across all robots. A true positive is when a robot that is near a novel object reports a novelty
numOfNoveltyFalsePositivesData = [];  # each data point is collated number of false positives in a certain pass and run, across all robots. A false positive is when a robot that is not near a novel object reports a novelty

robotsWinningNeuronErroValues = [];
robotsWinningNeuronNeigboursErroValues = [];
envChangedTimes = [];
networkTrainingTimes = [];

for p in range(len(PARAM_SETS)):
   config.dataDirectoryPath = DATA_DIRECTORY + "/" + PARAM_SETS[p];

   envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
   envChangedTimes.append(envChangedTimesData);
   averagePassLength, maxPassLength = helpers.calculatePassLength(config);
   noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);
   networkTrainingTimesData = gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime);
   networkTrainingTimes.append(networkTrainingTimesData);

   if (DO_NOVELTY_REPORTING_OVER_PASSES):
      numOfNoveltyTruePositivesData.append(gaData_module.getNumOfNoveltyTruePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));
      numOfNoveltyFalsePositivesData.append(gaData_module.getNumOfNoveltyFalsePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));

   if (DO_NEURON_ERRORS):
      config.outputDirectoryPath = OUTPUT_DIRECTORY + "/" + COMPARISON_NAME + "/";
      robotsWinningNeuronErrorValuesOverTime = gaData_module.getWinningNeuronQuantizationErrorOverTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);
      robotsWinningNeuronNeigboursErrorValuesOverTime = gaData_module.getWinningNeuronNeigboursQuantizationErrorOverTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);

      #-- plot over time
      yMax = 3;
      visualization.plotTimeSeries(config, [robotsWinningNeuronErrorValuesOverTime],"avgWinningNeuronError" + legendLabels[p] + ".png",constants.TEXT_QUANTIZATION_ERROR, yMax_=yMax);
      visualization.plotTimeSeries(config, [robotsWinningNeuronNeigboursErrorValuesOverTime],"avgWinningNeuronNeighbourError" + legendLabels[p] + ".png",constants.TEXT_QUANTIZATION_ERROR, yMax_=yMax);

      #-- get averages for later plot
      robotsWinningNeuronErroValues.append(robotsWinningNeuronErrorValuesOverTime);
      robotsWinningNeuronNeigboursErroValues.append(robotsWinningNeuronNeigboursErrorValuesOverTime);


#===================== PLOT =============================

if (DO_NOVELTY_REPORTING_OVER_PASSES):
   maxPassNo = 10;
   if (DO_BLOB_CAMERA_NOIZE):
      maxPassNo = 20;

   visualization.plotNoveltyReportingByPass(config, numOfNoveltyTruePositivesData, "true positives", legendLabels, yMax_=1.0, maxPassNo_=maxPassNo);

   brokenPlotYMax = crGraphs.INVALID_VALUE;
   brokenPlotYCoord = crGraphs.INVALID_VALUE;
   maxPassNo = 10;

   if (DO_BLOB_CAMERA_NOIZE):
      brokenPlotYMax = 40;
      brokenPlotYCoord = 1.0;
      maxPassNo = 20;
   visualization.plotNoveltyReportingByPass(config, numOfNoveltyFalsePositivesData, "false positives", legendLabels, brokenLinePlot_yMax_=brokenPlotYMax,brokenLinePlot_yCoord=brokenPlotYCoord, maxPassNo_=maxPassNo);

if (DO_NEURON_ERRORS):

   visualization.plotNetworkQuantizationErrors(config, envChangedTimes, networkTrainingTimes, robotsWinningNeuronErroValues, robotsWinningNeuronNeigboursErroValues, legendLabels);


   #filePath = OUTPUT_DIRECTORY + "/" + COMPARISON_NAME + "/avgWinningNeuronError.png";
   #crGraphs.createLinePlot([averageWinningNeuronErrorData], filePath_=filePath, xTickLabels_=legendLabels, yMin_=0, yMax_=2);

   #filePath = OUTPUT_DIRECTORY + "/" + COMPARISON_NAME + "/avgWinningNeuronNeighbourError.png";
   #crGraphs.createLinePlot([averageWinningNeuronNeighbourErrorData], filePath_=filePath, xTickLabels_=legendLabels, yMin_=0, yMax_=2);


#===================== WRITE PARAMETER DESCRIPTIONS =============================
parameterDescriptions = ""
for p in range(len(PARAM_SETS)):
   parameterDescriptions += "(" + LETTERS[p] + ")"  + " " + PARAM_SETS[p] + " ";

f = open(config.outputDirectoryPath + "/parameterSets.txt",'w')
f.write(parameterDescriptions)
f.close()

