####################################################################
##
## Code for the analysis of the impact of habituation update tau and alpha
## parameters
##
## Author: L. Pitonakova (http://lenkaspace.net)
## License: GNU General Public License. Please credit me when using my work.
##
####################################################################

import gaData_module;
from pyCreeper import crGraphs;
from pyCreeper import crGraphStyle;
from pyCreeper import crData;
import numpy;
import helpers;
import constants;
import visualization;
import paths;

config = visualization.PlotConfig()

ANALYSIS_NAME = "test";
ANALYSIS_NAME = "s1_x1_corridor";
#ANALYSIS_NAME = "s1_x1_room";

#-- define and create directories
DATA_DIRECTORY = paths.BASE_FILE_PATH_DATA + "/" + ANALYSIS_NAME;
OUTPUT_DIRECTORY = paths.BASE_FILE_PATH_OUTPUT_IMAGES + "/" + ANALYSIS_NAME + "_hut-hua_thr" + str(config.noveltyValueThreshold);
helpers.createFolder(OUTPUT_DIRECTORY)

#-- select graphs
DO_TIME_SERIES_PLOTS_NOVELTY = False;
DO_TIME_SERIES_PLOTS_NETWORK = False;
DO_SINGLE_MEASURE_PLOTS = True;


#-- define parameters
if ("corridor" in ANALYSIS_NAME):
   INW_VAL = 12;
   AT_VAL = 0.8;
else:
   INW_VAL = 12;
   AT_VAL = 0.7;

HABITUATION_UPDATE_ALPHA_VALS = [1.00, 1.05, 1.10, 1.20, 1.30];
WINNING_NEURON_HABITUATION_UPDATE_TAU_VALS = [0.66, 1.33, 2.00, 2.66, 3.33, 4.00, 5.00, 7.00, 9.00];

#HABITUATION_UPDATE_ALPHA_VALS = [1.30];



config.numOfRuns = 50;
config.numOfRobots = 1;



#--------------------------

initialNetworkTrainingTimeData = [[] for inw in range(len(HABITUATION_UPDATE_ALPHA_VALS))]
afterEngChangeNetworkTrainingTimeData = [[] for inw in range(len(HABITUATION_UPDATE_ALPHA_VALS))];

numOfRunsNovelObjectReportedData = [[] for inw in range(len(HABITUATION_UPDATE_ALPHA_VALS))];  # each data point is number of runs in a certain pass, in which a novelty object was reported while robot was near it.

numOfNoveltyTruePositivesData = [[] for inw in range(len(HABITUATION_UPDATE_ALPHA_VALS))];      # each data point is collated number of true positives in a certain pass and run, across all robots. A true positive is when a robot that is near a novel object reports a novelty
numOfNoveltyFalsePositivesData = [[] for inw in range(len(HABITUATION_UPDATE_ALPHA_VALS))];    # each data point is collated number of false positives in a certain pass and run, across all robots. A false positive is when a robot that is not near a novel object reports a novelty
numOfNoveltyFalseNegativesData = [[] for inw in range(len(HABITUATION_UPDATE_ALPHA_VALS))];    # each data point is collated number of false negatives in a certain pass and run, across all robots. A false negative is when a robot that is near a novel object does not report a novelty

numOfNeuronsData = [[] for inw in range(len(HABITUATION_UPDATE_ALPHA_VALS))];  # each data point is maximum number of neurons a network reached, collated over runs
percentageUsedNeuronsData = [[] for inw in range(len(HABITUATION_UPDATE_ALPHA_VALS))];  # each data point is maximum number of neurons a network reached, collated over runs and time

networkGrowthRateData = [[] for inw in range(len(HABITUATION_UPDATE_ALPHA_VALS))];  # each data point is average growth rate (% size increase per second), collated over runs


for hua in range(len(HABITUATION_UPDATE_ALPHA_VALS)):
   huaVal = HABITUATION_UPDATE_ALPHA_VALS[hua];
   for hut in range(len(WINNING_NEURON_HABITUATION_UPDATE_TAU_VALS)):
      hutVal = WINNING_NEURON_HABITUATION_UPDATE_TAU_VALS[hut];
      if (ANALYSIS_NAME == "test"):
         config.dataDirectoryPath = DATA_DIRECTORY;
         config.outputDirectoryPath = OUTPUT_DIRECTORY;
      else:
         config.dataDirectoryPath = DATA_DIRECTORY + "/inw{}_at{}_hut{:.2f}_hua{:.2f}".format(INW_VAL, AT_VAL, hutVal, huaVal);
         config.outputDirectoryPath = OUTPUT_DIRECTORY + "/novelObjectReporting_timeSeries/inw{}_at{}_hut{:.2f}_hua{:.2f}".format(INW_VAL, AT_VAL, hutVal, huaVal);


      envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
      averagePassLength, maxPassLength = helpers.calculatePassLength(config);

      noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);

      if (DO_TIME_SERIES_PLOTS_NOVELTY):
         #-- get time series data
         robotsNoveltyValues = gaData_module.getNoveltyValues(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);
         robotsAroundNovelObjectTimes = gaData_module.getRobotsAroundNoveltyTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.robotNoveltyDetectionRange*2);
         robotsNoveltyValueByDistanceToNovelObjectData = gaData_module.getNoveltyValuesByDistanceToNovelty(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 5, config.robotNoveltyDetectionRange*2);

         #print(robotsAroundNovelObjectTimes)
         #-- render time series plots
         #print("================")
         visualization.plotTimeSeries(config, [robotsNoveltyValues],"avgNoveltyVal.png","Avg. novelty",1, timeEnd_=10*maxPassLength);
         visualization.plotNoveltyValueAroundNovelObject(config, robotsNoveltyValues, robotsAroundNovelObjectTimes, robotsNoveltyValueByDistanceToNovelObjectData, config.robotNoveltyDetectionRange*2);

      if (DO_TIME_SERIES_PLOTS_NETWORK):

         robotsNetworkSize = gaData_module.getNumOfNeuronsOverTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);
         robotsNumOfUsedNeurons = gaData_module.getNumOfUsedNeuronsOverTimeSince(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 0);
         robotsNumOfUsedNeuronsAfterTraining = gaData_module.getNumOfUsedNeuronsOverTimeSince(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, noveltyOccurrenceMinTime);

         config.outputDirectoryPath = OUTPUT_DIRECTORY + "/neuralNetwork_timeSeries/inw{}_at{}_hut{:.2f}_hua{:.2f}".format(INW_VAL, AT_VAL, hutVal, huaVal);

         visualization.plotTimeSeries(config, [robotsNetworkSize,robotsNumOfUsedNeurons, robotsNumOfUsedNeuronsAfterTraining],"networkNeuronsUsage.png",constants.TEXT_NUM_OF_NEURONS, crGraphs.INVALID_VALUE, legendLabels_=["Total neurons", "Used since the beginning", "Used since change"]);

      #-- get single measure data
      if (DO_SINGLE_MEASURE_PLOTS):

         initialNetworkTrainingTimeData[hua].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime));
         afterEngChangeNetworkTrainingTimeData[hua].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, noveltyOccurrenceMinTime, config.totalExperimentTime*10));

         numOfRunsNovelObjectReportedData[hua].append(gaData_module.getNumOfRunsNovelObjectReported(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));

         numOfNoveltyTruePositivesData[hua].append(gaData_module.getNumOfNoveltyTruePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));
         numOfNoveltyFalseNegativesData[hua].append(gaData_module.getNumOfNoveltyFalseNegatives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));
         numOfNoveltyFalsePositivesData[hua].append(gaData_module.getNumOfNoveltyFalsePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));

         numOfNeuronsData[hua].append(gaData_module.getNumOfNeurons(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         networkGrowthRateData[hua].append(gaData_module.getNetworkGrowthRate(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, noveltyOccurrenceMinTime));


if (DO_SINGLE_MEASURE_PLOTS):
   config.style.markers = ['bs','cs','gs','ms','ks','rs'];

   #-- generate tick labels for different parameters
   legendLabels = [];
   for i in range(len(HABITUATION_UPDATE_ALPHA_VALS)):
      legendLabels.append(str(HABITUATION_UPDATE_ALPHA_VALS[i]));

   #-- render single-measure plots
   config.outputDirectoryPath = OUTPUT_DIRECTORY + "/"
   xAxisLabel = "Winning neuron habituation update tau";
   xAxisTicks = WINNING_NEURON_HABITUATION_UPDATE_TAU_VALS;

   #visualization.plotNetworkTrainingTime(config, initialNetworkTrainingTimeData, 0, noveltyOccurrenceMinTime, legendLabels, xAxisLabel, xAxisTicks);
   #visualization.plotNetworkTrainingTime(config, afterEngChangeNetworkTrainingTimeData, noveltyOccurrenceMinTime, "end", legendLabels, xAxisLabel, xAxisTicks);

   #visualization.plotNumOfRunsNovelObjectReported(config, numOfRunsNovelObjectReportedData, legendLabels, xAxisLabel, xAxisTicks);

   #visualization.plotNoveltyReporting(config, numOfNoveltyTruePositivesData, "true positives", legendLabels, xAxisLabel, xAxisTicks, yMax_=1.0);
   #visualization.plotNoveltyReporting(config, numOfNoveltyFalseNegativesData, "false negatives", legendLabels, xAxisLabel, xAxisTicks);
   #visualization.plotNoveltyReporting(config, numOfNoveltyFalsePositivesData, "false positives", legendLabels, xAxisLabel, xAxisTicks, brokenLinePlot_yMax_=6, brokenLinePlot_yCoord=1);

   visualization.plotNNMetric(config, numOfNeuronsData, "numOfNeurons", constants.TEXT_MAX_NUM_OF_NEURONS, legendLabels, xAxisLabel, xAxisTicks);

   visualization.plotNNMetric(config, networkGrowthRateData, "networkGrowthRate", constants.TEXT_GROWTH_RATE, legendLabels, xAxisLabel, xAxisTicks);
