####################################################################
##
## Code for the analysis of the impact of the number of input weights
## and activation threshold in basic environments without noise and
## with pre-set robot speed
##
## Author: L. Pitonakova (http://lenkaspace.net)
## License: GNU General Public License. Please credit me when using my work.
##
####################################################################

import gaData_module;
from pyCreeper import crGraphs;
from pyCreeper import crGraphStyle;
from pyCreeper import crData;
import numpy;
import math;
import helpers;
import constants;
import visualization;
import paths;

ANALYSIS_NAME = "test";
ANALYSIS_NAME = "s1_x1_corridor";
ANALYSIS_NAME = "s1_x1_room";

HABITUATION_THRESHOLD = 0.30;

config = visualization.PlotConfig()

#-- define and create directories
DATA_DIRECTORY = paths.BASE_FILE_PATH_DATA + "/" + ANALYSIS_NAME;
OUTPUT_DIRECTORY = paths.BASE_FILE_PATH_OUTPUT_IMAGES + "/" + ANALYSIS_NAME + "_inw-at_ht" + str(HABITUATION_THRESHOLD) + "_thr" + str(config.noveltyValueThreshold);
helpers.createFolder(OUTPUT_DIRECTORY)

#-- select graphs
DO_TIME_SERIES_PLOTS_NOVELTY = False;
DO_TIME_SERIES_PLOTS_NETWORK = True;
DO_SINGLE_MEASURE_PLOTS = False;
DO_HISTOGRAMS_NETWORK = False;

#-- define parameters
ACTIVATION_THRESHOLD_VALS = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

IN_WEIGHT_VALS = [9, 12, 18, 24, 30, 36, 60, 84, 39, "Adaptive"];
if ("room" in ANALYSIS_NAME):
   IN_WEIGHT_VALS = [9, 12, 18, 24, 36, 39, 60, 84, 30, "Adaptive"];

if (DO_HISTOGRAMS_NETWORK or DO_TIME_SERIES_PLOTS_NETWORK or DO_TIME_SERIES_PLOTS_NOVELTY):
   IN_WEIGHT_VALS = [IN_WEIGHT_VALS[-2], "Adaptive"];
   ACTIVATION_THRESHOLD_VALS = [0.2];
   IN_WEIGHT_VALS = [9];
   ACTIVATION_THRESHOLD_VALS = [0.7];

#IN_WEIGHT_VALS = [IN_WEIGHT_VALS[-2],"Adaptive"];
#ACTIVATION_THRESHOLD_VALS = [0.1,0.2];

#IN_WEIGHT_VALS = ["Adaptive"];


#ACTIVATION_THRESHOLD_VALS = [0.8];
#ACTIVATION_THRESHOLD_VALS = [0.3]
#IN_WEIGHT_VALS = ["Adaptive"];

config.numOfRuns = 50;
config.numOfRobots = 1;

#--------------------------

averagePassLengths = [[] for p in range(len(IN_WEIGHT_VALS))];
noveltyOccurrenceMinTimes = [[] for p in range(len(IN_WEIGHT_VALS))];

initialNetworkTrainingTimeData = [[] for inw in range(len(IN_WEIGHT_VALS))]
afterEngChangeNetworkTrainingTimeData = [[] for inw in range(len(IN_WEIGHT_VALS))];

numOfRunsNovelObjectReportedData = [[] for inw in range(len(IN_WEIGHT_VALS))];  # each data point is number of runs in a certain pass, in which a novelty object was reported while robot was near it.

numOfNoveltyTruePositivesData = [[] for inw in range(len(IN_WEIGHT_VALS))];      # each data point is collated number of true positives in a certain pass and run, across all robots. A true positive is when a robot that is near a novel object reports a novelty
numOfNoveltyFalsePositivesData = [[] for inw in range(len(IN_WEIGHT_VALS))];    # each data point is collated number of false positives in a certain pass and run, across all robots. A false positive is when a robot that is not near a novel object reports a novelty
numOfNoveltyFalseNegativesData = [[] for inw in range(len(IN_WEIGHT_VALS))];    # each data point is collated number of false negatives in a certain pass and run, across all robots. A false negative is when a robot that is near a novel object does not report a novelty

numOfNeuronsData = [[] for inw in range(len(IN_WEIGHT_VALS))];  # each data point is maximum number of neurons a network reached, collated over runs
numOfInWeightsData = [[] for inw in range(len(IN_WEIGHT_VALS))];  # each data point is average total number of neuron input weights a network had, collated over runs
numOfInWeightsPerNeuronData = [[] for inw in range(len(IN_WEIGHT_VALS))];  # each data point is average number of neuron input weights per neuron, collated over runs
numOfConnectionsData = [[] for inw in range(len(IN_WEIGHT_VALS))];  # each data point is average total number of inter-neuron connections a network had, collated over runs
percentageUsedNeuronsData = [[] for inw in range(len(IN_WEIGHT_VALS))];  # each data point is maximum number of neurons a network reached, collated over runs and time
networkGrowthRateData = [[] for inw in range(len(IN_WEIGHT_VALS))];  # each data point is average growth rate (% size increase per second), collated over runs

for inw in range(len(IN_WEIGHT_VALS)):
   inwVal = IN_WEIGHT_VALS[inw];
   for at in range(len(ACTIVATION_THRESHOLD_VALS)):
      atVal = ACTIVATION_THRESHOLD_VALS[at];
      if (ANALYSIS_NAME == "test"):
         config.dataDirectoryPath = DATA_DIRECTORY;
         config.outputDirectoryPath = OUTPUT_DIRECTORY;#
      else:
         if (HABITUATION_THRESHOLD == 0.15):
            config.dataDirectoryPath = DATA_DIRECTORY + "/inw" + str(inwVal) + "_at" + str(atVal);
         elif (HABITUATION_THRESHOLD == 0.30):
            config.dataDirectoryPath = paths.BASE_FILE_PATH_DATA + "/" + ANALYSIS_NAME.replace("x1","x2") + "/inw" + str(inwVal) + "_at" + str(atVal) + "_rms20";
         config.outputDirectoryPath = OUTPUT_DIRECTORY + "/novelObjectReporting_timeSeries/inw" + str(inwVal) + "_at" + str(atVal);


      envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
      averagePassLength, maxPassLength = helpers.calculatePassLength(config);
      averagePassLengths[inw].append(averagePassLength);

      noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);
      noveltyOccurrenceMinTimes[inw].append(noveltyOccurrenceMinTime);

      if (DO_TIME_SERIES_PLOTS_NOVELTY):
         #-- get time series data
         robotsNoveltyValues = gaData_module.getNoveltyValues(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);
         robotsAroundNovelObjectTimes = gaData_module.getRobotsAroundNoveltyTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.robotNoveltyDetectionRange*2);
         robotsNoveltyValueByDistanceToNovelObjectData = gaData_module.getNoveltyValuesByDistanceToNovelty(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 5, config.robotNoveltyDetectionRange*2);

         #print(robotsAroundNovelObjectTimes)
         #-- render time series plots
         #print("================")
         visualization.plotTimeSeries(config, [robotsNoveltyValues],"avgNoveltyVal.png", constants.TEXT_AVERAGE_NOVELTY,1, timeEnd_=10*maxPassLength);
         visualization.plotTimeSeries(config, [robotsNoveltyValues],"avgNoveltyVal_aroundChange.png",constants.TEXT_AVERAGE_NOVELTY,1, timeStart_=19*maxPassLength, timeEnd_=21*maxPassLength);
         visualization.plotNoveltyValueAroundNovelObject(config, robotsNoveltyValues, robotsAroundNovelObjectTimes, robotsNoveltyValueByDistanceToNovelObjectData, config.robotNoveltyDetectionRange*2);

      if (DO_TIME_SERIES_PLOTS_NETWORK):

         robotsNetworkSize = gaData_module.getNumOfNeuronsOverTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);
         robotsNumOfUsedNeurons = gaData_module.getNumOfUsedNeuronsOverTimeSince(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 0);
         robotsNumOfUsedNeuronsAfterTraining = gaData_module.getNumOfUsedNeuronsOverTimeSince(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, noveltyOccurrenceMinTime);

         config.outputDirectoryPath = OUTPUT_DIRECTORY + "/neuralNetwork_timeSeries/inw" + str(inwVal) + "_at" + str(atVal);
         yMax = crGraphs.INVALID_VALUE;
         if ("corridor" in ANALYSIS_NAME):
            yMax = 350;
            if ("at0.2" in config.dataDirectoryPath):
               yMax = 110;
         visualization.plotTimeSeries(config, [robotsNetworkSize,robotsNumOfUsedNeurons, robotsNumOfUsedNeuronsAfterTraining],"networkNeuronsUsage.png",constants.TEXT_NUM_OF_NEURONS, yMax, legendLabels_=["Total neurons", "Used since the beginning", "Used since change"]);

         visualization.plotNeuronUsageTimeSeries(config, robotsNetworkSize, robotsNumOfUsedNeurons, robotsNumOfUsedNeuronsAfterTraining);
      #-- get single measure data
      if (DO_SINGLE_MEASURE_PLOTS):

         initialNetworkTrainingTimeData[inw].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime));
         afterEngChangeNetworkTrainingTimeData[inw].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, noveltyOccurrenceMinTime, config.totalExperimentTime*10));

         numOfRunsNovelObjectReportedData[inw].append(gaData_module.getNumOfRunsNovelObjectReported(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));

         numOfNoveltyTruePositivesData[inw].append(gaData_module.getNumOfNoveltyTruePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));
         #numOfNoveltyFalseNegativesData[inw].append(gaData_module.getNumOfNoveltyFalseNegatives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));
         numOfNoveltyFalsePositivesData[inw].append(gaData_module.getNumOfNoveltyFalsePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));

         numOfNeuronsData[inw].append(gaData_module.getNumOfNeurons(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         numOfInWeightsData[inw].append(gaData_module.getNumOfInWeights(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         numOfInWeightsPerNeuronData[inw].append(gaData_module.getNumOfInWeightsPerNeuron(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         numOfConnectionsData[inw].append(gaData_module.getNumOfConnections(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         networkGrowthRateData[inw].append(gaData_module.getNetworkGrowthRate(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, noveltyOccurrenceMinTime));

      #------------------------------------
      # network usage
      #------------------------------------
      if (DO_HISTOGRAMS_NETWORK):
         networkTrainingTimeAverage = (int)(math.ceil(crData.getAverageOfAList(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime))));
         config.outputDirectoryPath = OUTPUT_DIRECTORY + "/neuralNetwork_histograms/inw" + str(inwVal) + "_at" + str(atVal);

         if ("corridor" in ANALYSIS_NAME):
            yMax = 7;
            if (atVal == 0.2):
               yMax = 3;
         else:
            yMax = 4;
            if (atVal == 0.2):
               yMax = 1.4
         yMax = crGraphs.INVALID_VALUE;
         neuronUsageInBins = gaData_module.getNeuronUsageAsBins(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,0,networkTrainingTimeAverage); #networkTrainingTimeAverage
         fileName = "neuronUsage_beforeNovelty_training_t0-{}".format(networkTrainingTimeAverage);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, firstNBins_=50, yMax_=yMax);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, lastNBins_=20);

         if ("corridor" in ANALYSIS_NAME):
            yMax = 30;
            if (atVal == 0.2):
               yMax = 1.4;
         else:
            yMax = 10;
            if (atVal == 0.2):
               yMax = 0.16
         yMax = crGraphs.INVALID_VALUE;
         neuronUsageInBins = gaData_module.getNeuronUsageAsBins(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,networkTrainingTimeAverage,noveltyOccurrenceMinTime);
         fileName = "neuronUsage_beforeNovelty_trained_t{}-{}".format(networkTrainingTimeAverage,noveltyOccurrenceMinTime);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, firstNBins_=50, yMax_=yMax);

         if ("corridor" in ANALYSIS_NAME):
            yMax = 60;
            if (atVal == 0.2):
               yMax = 12;
         else:
            yMax = 12;
            if (atVal == 0.2):
               yMax = 2.5;
         yMax = crGraphs.INVALID_VALUE;
         neuronUsageInBins = gaData_module.getNeuronUsageAsBins(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,networkTrainingTimeAverage,(int)(networkTrainingTimeAverage + averagePassLength));
         fileName = "neuronUsage_beforeNovelty_trained_firstPass";
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, firstNBins_=50, yMax_=yMax);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, lastNBins_=20);



if (DO_SINGLE_MEASURE_PLOTS):
   config.style.markers = ['bs','cs','gs','ms','ks', 'bo', 'co', 'go', 'ro', 'rd'];
   config.style.lineStyles = ['-','-','-','-','-','-','-','-','-','--'];

   #-- render single-measure plots
   config.outputDirectoryPath = OUTPUT_DIRECTORY + "/"
   xAxisLabel = "Activation threshold";
   xAxisTicks = ACTIVATION_THRESHOLD_VALS;
   legendLabels = IN_WEIGHT_VALS;

   #------------------------------------
   # network training time
   #------------------------------------

   visualization.plotNetworkTrainingTime(config, initialNetworkTrainingTimeData, 0, noveltyOccurrenceMinTime, legendLabels, xAxisLabel, xAxisTicks);
   visualization.plotNetworkTrainingTime(config, afterEngChangeNetworkTrainingTimeData, noveltyOccurrenceMinTime, "end", legendLabels, xAxisLabel, xAxisTicks);

   visualization.plotNetworkTrainingTimeByPass(config, initialNetworkTrainingTimeData, averagePassLengths, [[0 for v in range(len(ACTIVATION_THRESHOLD_VALS))] for p in range(len(IN_WEIGHT_VALS))], legendLabels, xAxisLabel, xAxisTicks);
   visualization.plotNetworkTrainingTimeByPass(config, afterEngChangeNetworkTrainingTimeData, averagePassLengths, noveltyOccurrenceMinTimes, legendLabels, xAxisLabel, xAxisTicks, yMax_=30, startTimeLabel_="afterChange"); #y max can be more than 20, because the total time really its more than 40 passes


   #------------------------------------
   # reporting performance
   #------------------------------------

   #visualization.plotNumOfRunsNovelObjectReported(config, numOfRunsNovelObjectReportedData, legendLabels, xAxisLabel, xAxisTicks);

   visualization.plotNoveltyReporting(config, numOfNoveltyTruePositivesData, "true positives", legendLabels, xAxisLabel, xAxisTicks);
   #visualization.plotNoveltyReporting(config, numOfNoveltyFalseNegativesData, "false negatives", legendLabels, xAxisLabel, xAxisTicks);

   brokenPlotYMax = 1.5;
   brokenPlotYCoord = 0.6;
   visualization.plotNoveltyReporting(config, numOfNoveltyFalsePositivesData, "false positives", legendLabels, xAxisLabel, xAxisTicks, brokenLinePlot_yMax_=brokenPlotYMax,brokenLinePlot_yCoord=brokenPlotYCoord);


   #------------------------------------
   # network size
   #------------------------------------

   brokenPlotYMax = 2000;
   brokenPlotYCoord = 100;
   visualization.plotNNMetric(config, numOfNeuronsData, "numOfNeurons", constants.TEXT_MAX_NUM_OF_NEURONS, legendLabels, xAxisLabel, xAxisTicks, brokenLinePlot_yMax_=brokenPlotYMax,brokenLinePlot_yCoord=brokenPlotYCoord);

   brokenPlotYMax = 90000;
   brokenPlotYCoord = 10000;
   if ("room" in ANALYSIS_NAME):
      brokenPlotYMax = 40000;
      brokenPlotYCoord = 4000;
   visualization.plotNNMetric(config, numOfInWeightsData, "numOfInWeights", constants.TEXT_AVERAGE_TOTAL_INPUT_WEIGHTS, legendLabels, xAxisLabel, xAxisTicks, brokenLinePlot_yMax_=brokenPlotYMax,brokenLinePlot_yCoord=brokenPlotYCoord);

   visualization.plotNNMetric(config, numOfInWeightsPerNeuronData, "numOfInWeightsPerNeuron", constants.TEXT_AVERAGE_INPUT_WEIGHTS_PER_NEURON, legendLabels, xAxisLabel, xAxisTicks);
   visualization.plotNNMetric(config, numOfConnectionsData, "numOfConnections", constants.TEXT_AVERAGE_TOTAL_CONNECTIONS, legendLabels, xAxisLabel, xAxisTicks);

   visualization.plotNNMetric(config, networkGrowthRateData, "networkGrowthRate", constants.TEXT_GROWTH_RATE, legendLabels, xAxisLabel, xAxisTicks);


