####################################################################
##
## Code for the analysis of the impact of various single parameters
## in basic environments without noise and with pre-set robot speed
##
## Author: L. Pitonakova (http://lenkaspace.net)
## License: GNU General Public License. Please credit me when using my work.
##
####################################################################

import gaData_module;
from pyCreeper import crGraphs;
from pyCreeper import crGraphStyle;
from pyCreeper import crData;
import numpy;
import helpers;
import constants;
import visualization;
import paths;

config = visualization.PlotConfig()

ANALYSIS_NAME = "s1_x1_corridor";
#ANALYSIS_NAME = "s1_x1_room";


#-- select type of graph
DO_TIME_SERIES_PLOTS_NOVELTY = False;
DO_TIME_SERIES_PLOTS_NETWORK = True;
DO_SINGLE_MEASURE_PLOTS = False;

#-- define parameter combinations, set in the INW-AT.sh experiment
if ("corridor" in ANALYSIS_NAME):
   IN_WEIGHT_ACTIVATION_THRESHOLD_SETS = [[12,0.8],[39,0.2],["Adaptive",0.2]];
else:
   IN_WEIGHT_ACTIVATION_THRESHOLD_SETS = [[9,0.7],[30,0.2],["Adaptive",0.2]];

#-- select parameter
DO_HABITUATION_THRESHOLD = False;
DO_WINNING_NEURON_LEARNING_RATE = True;
DO_UPDATES_NEIGHBOUR_NEURON_PROPORTIONALITY = False;
DO_MAXIMUM_CONNECTION_AGE = False;
DO_DISHABITUATION = False;

config.numOfRuns = 50;
config.numOfRobots = 1;


PARAM_FORMAT = "{:.2f}";
if (DO_HABITUATION_THRESHOLD):
   PARAM_SHORTCUT = "ht";
   PARAM_NAME = constants.TEXT_HABITUATION_THRESHOLD;
   PARAM_VALS = [0.00, 0.05, 0.10, 0.15, 0.20, 0.25, 0.30, 0.40, 0.50, 0.60, 0.70, 0.80, 0.90, 1.00];
   if (DO_TIME_SERIES_PLOTS_NETWORK):
      PARAM_VALS = [0.30];
   X_TICK_VALS = PARAM_VALS;
elif (DO_WINNING_NEURON_LEARNING_RATE):
   PARAM_SHORTCUT = "lr";
   PARAM_NAME = constants.TEXT_LEARNING_RATE;
   PARAM_VALS = [0.00, 0.05, 0.10, 0.15, 0.20, 0.25, 0.30, 0.40, 0.50, 0.60, 0.70, 0.80, 0.90, 1.00];
   if (DO_TIME_SERIES_PLOTS_NETWORK):
      PARAM_VALS = [0.30, 1.0];

   if (DO_TIME_SERIES_PLOTS_NOVELTY):
      PARAM_VALS = [0.00, 0.30];
      if ("corridor" in ANALYSIS_NAME):
         IN_WEIGHT_ACTIVATION_THRESHOLD_SETS = [[12,0.8]];
      else:
         IN_WEIGHT_ACTIVATION_THRESHOLD_SETS = [[12,0.7]];
   X_TICK_VALS = PARAM_VALS;

elif (DO_UPDATES_NEIGHBOUR_NEURON_PROPORTIONALITY):
   PARAM_SHORTCUT = "unnp";
   PARAM_NAME = "Neighbour neuron updates proportionality"
   PARAM_VALS = [0.00, 0.01, 0.05, 0.10, 0.15, 0.20, 0.30, 0.40, 0.50, 0.55, 0.60, 0.70];
   if (DO_TIME_SERIES_PLOTS_NETWORK):
      PARAM_VALS = [0.01, 0.55];
   X_TICK_VALS = PARAM_VALS;

elif (DO_MAXIMUM_CONNECTION_AGE):
   PARAM_SHORTCUT = "mca";
   PARAM_NAME = "Maximum connection age"
   PARAM_VALS = [1, 3, 5, 7, 10, 15, 20, 30, 40, 50, 0];
   X_TICK_VALS = [1, 3, 5, 7, 10, 15, 20, 30, 40, 50, "-"];
   if (DO_TIME_SERIES_PLOTS_NETWORK):
      PARAM_VALS = [0,1,5];

elif (DO_DISHABITUATION):
   PARAM_SHORTCUT = "dishab_unnp";
   PARAM_NAME = "Neighbour neuron updates proportionality"
   PARAM_FORMAT = "{:.3f}";
   PARAM_VALS = [0.00, 0.001, 0.002, 0.003, 0.004, 0.005, 0.007, 0.01, 0.05, 0.10, 0.15];
   if (DO_TIME_SERIES_PLOTS_NETWORK or DO_TIME_SERIES_PLOTS_NOVELTY):
      PARAM_VALS = [0.00, 0.005];
   X_TICK_VALS = PARAM_VALS;


#-- define and create directories
DATA_DIRECTORY = paths.BASE_FILE_PATH_DATA + "/" + ANALYSIS_NAME;
OUTPUT_DIRECTORY = paths.BASE_FILE_PATH_OUTPUT_IMAGES + "/" + ANALYSIS_NAME + "_" + PARAM_SHORTCUT + "_thr" + str(config.noveltyValueThreshold);
helpers.createFolder(OUTPUT_DIRECTORY)


#IN_WEIGHT_ACTIVATION_THRESHOLD_SETS = [[12, 0.8]];
#PARAM_VALS = [0.30, 1.00];
X_TICK_VALS = PARAM_VALS;




#--------------------------

averagePassLengths = [[] for p in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];
noveltyOccurrenceMinTimes = [[] for p in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];

initialNetworkTrainingTimeData = [[] for inw in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))]
afterEngChangeNetworkTrainingTimeData = [[] for inw in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];

numOfRunsNovelObjectReportedData = [[] for inw in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];  # each data point is number of runs in a certain pass, in which a novelty object was reported while robot was near it.

numOfNoveltyTruePositivesData = [[] for inw in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];      # each data point is collated number of true positives in a certain pass and run, across all robots. A true positive is when a robot that is near a novel object reports a novelty
numOfNoveltyFalsePositivesData = [[] for inw in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];    # each data point is collated number of false positives in a certain pass and run, across all robots. A false positive is when a robot that is not near a novel object reports a novelty
numOfNoveltyFalseNegativesData = [[] for inw in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];    # each data point is collated number of false negatives in a certain pass and run, across all robots. A false negative is when a robot that is near a novel object does not report a novelty


numOfNeuronsData = [[] for inw in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];  # each data point is maximum number of neurons a network reached, collated over runs
percentageUsedNeuronsData = [[] for inw in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];  # each data point is maximum number of neurons a network reached, collated over runs and time
networkGrowthRateData = [[] for inw in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))];  # each data point is average growth rate (% size increase per second), collated over runs

for s in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS)):
   inwVal = IN_WEIGHT_ACTIVATION_THRESHOLD_SETS[s][0];
   atVal = IN_WEIGHT_ACTIVATION_THRESHOLD_SETS[s][1];

   for p in range(len(PARAM_VALS)):
      paramVal = PARAM_VALS[p];
      if (ANALYSIS_NAME == "test"):
         config.dataDirectoryPath = DATA_DIRECTORY;
         config.outputDirectoryPath = OUTPUT_DIRECTORY;
      else:
         config.dataDirectoryPath = DATA_DIRECTORY + ("/inw{}_at{}_{}" + PARAM_FORMAT).format(inwVal, atVal, PARAM_SHORTCUT, paramVal);
         config.outputDirectoryPath = OUTPUT_DIRECTORY + ("/novelObjectReporting_timeSeries/inw{}_at{}_{}" + PARAM_FORMAT).format(inwVal, atVal, PARAM_SHORTCUT, paramVal);


      envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
      averagePassLength, maxPassLength = helpers.calculatePassLength(config);
      averagePassLengths[s].append(averagePassLength);

      noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);
      noveltyOccurrenceMinTimes[s].append(noveltyOccurrenceMinTime);

      if (DO_TIME_SERIES_PLOTS_NOVELTY):
         #-- get time series data
         robotsNoveltyValues = gaData_module.getNoveltyValues(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);
         robotsAroundNovelObjectTimes = gaData_module.getRobotsAroundNoveltyTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.robotNoveltyDetectionRange*2);
         robotsNoveltyValueByDistanceToNovelObjectData = gaData_module.getNoveltyValuesByDistanceToNovelty(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 5, config.robotNoveltyDetectionRange*2);

         #print(robotsAroundNovelObjectTimes)
         #-- render time series plots
         #print("================")
         visualization.plotTimeSeries(config, [robotsNoveltyValues],"avgNoveltyVal.png",constants.TEXT_AVERAGE_NOVELTY,1, timeEnd_=10*maxPassLength);
         visualization.plotNoveltyValueAroundNovelObject(config, robotsNoveltyValues, robotsAroundNovelObjectTimes, robotsNoveltyValueByDistanceToNovelObjectData, config.robotNoveltyDetectionRange*2);

      if (DO_TIME_SERIES_PLOTS_NETWORK):

         robotsNetworkSize = gaData_module.getNumOfNeuronsOverTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);
         robotsNumOfUsedNeurons = gaData_module.getNumOfUsedNeuronsOverTimeSince(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 0);
         robotsNumOfUsedNeuronsAfterTraining = gaData_module.getNumOfUsedNeuronsOverTimeSince(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, noveltyOccurrenceMinTime);

         config.outputDirectoryPath = OUTPUT_DIRECTORY + ("/neuralNetwork_timeSeries/inw{}_at{}_{}" + PARAM_FORMAT).format(inwVal, atVal, PARAM_SHORTCUT, paramVal);

         #[robotsNetworkSize,robotsNumOfUsedNeurons, robotsNumOfUsedNeuronsAfterTraining]
         visualization.plotTimeSeries(config, [robotsNetworkSize, robotsNumOfUsedNeurons, robotsNumOfUsedNeuronsAfterTraining],"networkNeuronsUsage.png",constants.TEXT_NUM_OF_NEURONS, crGraphs.INVALID_VALUE, legendLabels_=["Total neurons", "Used since the beginning", "Used since change"]);
         visualization.plotNeuronUsageTimeSeries(config, robotsNetworkSize, robotsNumOfUsedNeurons, robotsNumOfUsedNeuronsAfterTraining);

      #-- get single measure data
      if (DO_SINGLE_MEASURE_PLOTS):

         initialNetworkTrainingTimeData[s].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime));
         afterEngChangeNetworkTrainingTimeData[s].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, noveltyOccurrenceMinTime, config.totalExperimentTime*10));

         numOfRunsNovelObjectReportedData[s].append(gaData_module.getNumOfRunsNovelObjectReported(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));

         numOfNoveltyTruePositivesData[s].append(gaData_module.getNumOfNoveltyTruePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));
         numOfNoveltyFalseNegativesData[s].append(gaData_module.getNumOfNoveltyFalseNegatives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));
         numOfNoveltyFalsePositivesData[s].append(gaData_module.getNumOfNoveltyFalsePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));

         numOfNeuronsData[s].append(gaData_module.getNumOfNeurons(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         networkGrowthRateData[s].append(gaData_module.getNetworkGrowthRate(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, noveltyOccurrenceMinTime));

         #print(networkGrowthRateData[s]);




if (DO_SINGLE_MEASURE_PLOTS):
   config.style.markers = ['cs','bs','rs'];
   config.style.lineStyles = ['-','-','--'];
   #-- generate tick labels for different parameters
   legendLabels = [];
   for s in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS)):
      inwVal = IN_WEIGHT_ACTIVATION_THRESHOLD_SETS[s][0];
      atVal = IN_WEIGHT_ACTIVATION_THRESHOLD_SETS[s][1];
      legendLabels.append('$||w_c||$={}, $\Theta_A$={}'.format(inwVal, atVal))

   config.outputDirectoryPath = OUTPUT_DIRECTORY + "/"
   xAxisLabel = PARAM_NAME;
   xAxisTicks = X_TICK_VALS;

   #------------------------------------
   # network training time
   #------------------------------------

   visualization.plotNetworkTrainingTime(config, initialNetworkTrainingTimeData, 0, noveltyOccurrenceMinTime, legendLabels, xAxisLabel, xAxisTicks);
   visualization.plotNetworkTrainingTime(config, afterEngChangeNetworkTrainingTimeData, noveltyOccurrenceMinTime, "end", legendLabels, xAxisLabel, xAxisTicks);

   visualization.plotNetworkTrainingTimeByPass(config, initialNetworkTrainingTimeData, averagePassLengths, [[0 for v in range(len(PARAM_VALS))] for p in range(len(IN_WEIGHT_ACTIVATION_THRESHOLD_SETS))], legendLabels, xAxisLabel, xAxisTicks);
   visualization.plotNetworkTrainingTimeByPass(config, afterEngChangeNetworkTrainingTimeData, averagePassLengths, noveltyOccurrenceMinTimes, legendLabels, xAxisLabel, xAxisTicks, yMax_=30, startTimeLabel_="afterChange"); #y max can be more than 20, because the total time really its more than 40 passes

   #--------------------------------
   #      NOVELTY METRICS
   #--------------------------------
   visualization.plotNoveltyReporting(config, numOfNoveltyTruePositivesData, "true positives", legendLabels, xAxisLabel, xAxisTicks, yMax_=1.0);

   brokenPlotYMax = crGraphs.INVALID_VALUE;
   brokenPlotYCoord = crGraphs.INVALID_VALUE;
   #if (DO_WINNING_NEURON_LEARNING_RATE):
   #   brokenPlotYMax = 0.9;
   #   brokenPlotYCoord = 0.1;
   if (DO_MAXIMUM_CONNECTION_AGE):
      brokenPlotYMax = 30;
      brokenPlotYCoord = 1.0;

   if (DO_DISHABITUATION):
      brokenPlotYCoord = 1.0;
      brokenPlotYMax = 70;
   visualization.plotNoveltyReporting(config, numOfNoveltyFalsePositivesData, "false positives", legendLabels, xAxisLabel, xAxisTicks, brokenLinePlot_yMax_=brokenPlotYMax, brokenLinePlot_yCoord=brokenPlotYCoord);




   #--------------------------------
   #      NETWORK METRICS
   #--------------------------------


   brokenPlotYMax = crGraphs.INVALID_VALUE;
   brokenPlotYCoord = crGraphs.INVALID_VALUE;
   yMax = crGraphs.INVALID_VALUE;
   if (DO_HABITUATION_THRESHOLD):
      #brokenPlotYMax = 1000;
      #brokenPlotYCoord = 400;
      if ("room" in ANALYSIS_NAME):
         brokenPlotYMax = 300;
         brokenPlotYCoord = 100;
   if (DO_DISHABITUATION):
      yMax = 600;
   visualization.plotNNMetric(config, numOfNeuronsData, "numOfNeurons", constants.TEXT_MAX_NUM_OF_NEURONS, legendLabels, xAxisLabel, xAxisTicks, yMax_=yMax, brokenLinePlot_yMax_=brokenPlotYMax, brokenLinePlot_yCoord=brokenPlotYCoord);

   brokenPlotYMax = crGraphs.INVALID_VALUE;
   brokenPlotYCoord = crGraphs.INVALID_VALUE;
   yMax = crGraphs.INVALID_VALUE;
   if (DO_HABITUATION_THRESHOLD or DO_WINNING_NEURON_LEARNING_RATE):
      brokenPlotYMax = 70;
      brokenPlotYCoord = 20;

   if (DO_MAXIMUM_CONNECTION_AGE):
      brokenPlotYCoord = 5;
      brokenPlotYMax = 60;

   if (DO_DISHABITUATION):
      brokenPlotYCoord = 10;
      brokenPlotYMax = 60;

   visualization.plotNNMetric(config, networkGrowthRateData, "networkGrowthRate", constants.TEXT_GROWTH_RATE, legendLabels, xAxisLabel, xAxisTicks, yMax_=yMax, brokenLinePlot_yCoord=brokenPlotYCoord, brokenLinePlot_yMax_=brokenPlotYMax);



