####################################################################
##
## Code for the analysis of the impact of various parameters such as
## robot speed and input noise on selected networks
##
## Author: L. Pitonakova (http://lenkaspace.net)
## License: GNU General Public License. Please credit me when using my work.
##
#####################################################################

#--------------------------------
#      ENVIRONMENT METRICS - comparing different sets
#--------------------------------

import gaData_module;
from string import ascii_lowercase
from pyCreeper import crGraphs;
from pyCreeper import crGraphStyle;
from pyCreeper import crData;
import numpy;
import math;
import helpers;
import constants;
import visualization;
import paths;


config = visualization.PlotConfig()

ANALYSIS_NAME = "s1_x2_corridor";
ANALYSIS_NAME = "s1_x2_room";



if ("corridor" in ANALYSIS_NAME):
   PARAM_SETS = ["inw12_at0.8", "inw39_at0.2", "inwAdaptive_at0.2"];
else:
   PARAM_SETS = ["inw9_at0.7", "inw30_at0.2", "inwAdaptive_at0.2"];

DO_MAX_ROBOT_SPEED = True;
DO_BLOB_CAMERA_NOIZE = False;

DO_TIME_SERIES_PLOTS_NETWORK = False;
DO_SINGLE_MEASURE_PLOTS = True;
DO_HISTOGRAMS_NETWORK = False;

config.numOfRuns = 50;

#-- define parameter sets
if (DO_MAX_ROBOT_SPEED):
   COMPARISON_NAME = "maxRobotSpeed"
   PARAM_SHORTCUT = "rms";
   PARAM_NAME = constants.TEXT_MAX_ROBOT_SPEED;
   PARAM_VALS = [5, 10, 15, 20, 25, 30, 40, 60, 80];

   if (DO_TIME_SERIES_PLOTS_NETWORK or DO_HISTOGRAMS_NETWORK):
      PARAM_VALS = [5, 20, 40];
   X_TICK_VALS = PARAM_VALS;

if (DO_BLOB_CAMERA_NOIZE):
   COMPARISON_NAME = "blobCameraNoize"
   PARAM_SHORTCUT = "bcn";
   PARAM_NAME = constants.TEXT_BLOB_CAMERA_NOIZE;
   PARAM_VALS = [0, 0.01, 0.025, 0.05, 0.10, 0.15, 0.20, 0.30, 0.40];

   if (DO_TIME_SERIES_PLOTS_NETWORK or DO_HISTOGRAMS_NETWORK):
      PARAM_VALS = [0.05, 0.30];
   X_TICK_VALS = PARAM_VALS;

ROBOT_SPEED = 20;

#PARAM_SETS = ["inwAdaptive_at0.2"];
#PARAM_SETS = ["inw36_at0.2", "inwAdaptive_at0.2"];
#PARAM_VALS = [10, 20, 30, 60, 80];
X_TICK_VALS = PARAM_VALS;

#-- define and create directories
DATA_DIRECTORY = paths.BASE_FILE_PATH_DATA + "/" + ANALYSIS_NAME;
OUTPUT_DIRECTORY = paths.BASE_FILE_PATH_OUTPUT_IMAGES + "/" + ANALYSIS_NAME + "_" + COMPARISON_NAME + "_thr" + str(config.noveltyValueThreshold);
helpers.createFolder(OUTPUT_DIRECTORY)


#-- get data
averagePassLengths = [[] for p in range(len(PARAM_SETS))];
noveltyOccurrenceMinTimes = [[] for p in range(len(PARAM_SETS))];

initialNetworkTrainingTimeData = [[] for inw in range(len(PARAM_SETS))]
afterEngChangeNetworkTrainingTimeData = [[] for inw in range(len(PARAM_SETS))];

numOfNoveltyTruePositivesData = [[] for inw in range(len(PARAM_SETS))];      # each data point is collated number of true positives in a certain pass and run, across all robots. A true positive is when a robot that is near a novel object reports a novelty
numOfNoveltyFalsePositivesData = [[] for inw in range(len(PARAM_SETS))];    # each data point is collated number of false positives in a certain pass and run, across all robots. A false positive is when a robot that is not near a novel object reports a novelty

numOfNeuronsData = [[] for inw in range(len(PARAM_SETS))];  # each data point is maximum number of neurons a network reached, collated over runs
numOfInWeightsData = [[] for inw in range(len(PARAM_SETS))];  # each data point is average total number of neuron input weights a network had, collated over runs
numOfInWeightsPerNeuronData = [[] for inw in range(len(PARAM_SETS))];  # each data point is average number of neuron input weights per neuron, collated over runs
numOfConnectionsData = [[] for inw in range(len(PARAM_SETS))];  # each data point is average total number of inter-neuron connections a network had, collated over runs
networkGrowthRateData = [[] for inw in range(len(PARAM_SETS))];  # each data point is average growth rate (% size increase per second), collated over runs

neuronUsageInBins = [[] for inw in range(len(PARAM_SETS))];  # each data point is number of neurons that have fired as winning a certain number of times


for p in range(len(PARAM_SETS)):
   paramSetVal = PARAM_SETS[p];
   for v in range(len(PARAM_VALS)):
      subParamVal = PARAM_VALS[v];
      config.dataDirectoryPath = DATA_DIRECTORY + "/{}_{}{}".format(paramSetVal,PARAM_SHORTCUT,subParamVal)

      if (DO_BLOB_CAMERA_NOIZE):
         if (subParamVal == 0):
            #-- resuse data from robot speed experiments, where noize was 0
            config.dataDirectoryPath = DATA_DIRECTORY + "/{}_rms{}".format(paramSetVal,ROBOT_SPEED)
         else:
            config.dataDirectoryPath = DATA_DIRECTORY + "/{}_rms{}_{}{:.3f}".format(paramSetVal,ROBOT_SPEED,PARAM_SHORTCUT,subParamVal)

      averagePassLength, maxPassLength = helpers.calculatePassLength(config);
      averagePassLengths[p].append(averagePassLength);

      envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
      noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);
      noveltyOccurrenceMinTimes[p].append(noveltyOccurrenceMinTime);

      totalExperimentTime = helpers.calculateTotalExperimentTime(config);

      #------------------------------------
      # get single measure data
      #------------------------------------
      if (DO_SINGLE_MEASURE_PLOTS):

         initialNetworkTrainingTimeData[p].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime));
         afterEngChangeNetworkTrainingTimeData[p].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, noveltyOccurrenceMinTime, totalExperimentTime*10));

         numOfNoveltyTruePositivesData[p].append(gaData_module.getNumOfNoveltyTruePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));
         numOfNoveltyFalsePositivesData[p].append(gaData_module.getNumOfNoveltyFalsePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));

         numOfNeuronsData[p].append(gaData_module.getNumOfNeurons(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         numOfInWeightsData[p].append(gaData_module.getNumOfInWeights(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         numOfInWeightsPerNeuronData[p].append(gaData_module.getNumOfInWeightsPerNeuron(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         numOfConnectionsData[p].append(gaData_module.getNumOfConnections(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         networkGrowthRateData[p].append(gaData_module.getNetworkGrowthRate(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, noveltyOccurrenceMinTime));


      #------------------------------------
      # time series - network
      #------------------------------------
      if (DO_TIME_SERIES_PLOTS_NETWORK):

         robotsNetworkSize = gaData_module.getNumOfNeuronsOverTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);
         robotsNumOfUsedNeurons = gaData_module.getNumOfUsedNeuronsOverTimeSince(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 0);
         robotsNumOfUsedNeuronsAfterTraining = gaData_module.getNumOfUsedNeuronsOverTimeSince(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, noveltyOccurrenceMinTime);

         config.outputDirectoryPath = OUTPUT_DIRECTORY + "/neuralNetwork_timeSeries/{}_{}{}".format(paramSetVal,PARAM_SHORTCUT,subParamVal);

         yMax = crGraphs.INVALID_VALUE;
         if ("corridor" in ANALYSIS_NAME):
            if ("Adaptive" in config.outputDirectoryPath):
               yMax = 80;
            else:
               yMax = 350;
         visualization.plotTimeSeries(config, [robotsNetworkSize,robotsNumOfUsedNeurons, robotsNumOfUsedNeuronsAfterTraining],"networkNeuronsUsage.png",constants.TEXT_NUM_OF_NEURONS, legendLabels_=["Total neurons", "Used since the beginning", "Used since change"], yMax_=yMax);

      #------------------------------------
      # network usage
      #------------------------------------
      if (DO_HISTOGRAMS_NETWORK):
         networkTrainingTimeAverage = (int)(math.ceil(crData.getAverageOfAList(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime))));
         config.outputDirectoryPath = OUTPUT_DIRECTORY + "/neuralNetwork_histograms/{}_{}{}".format(paramSetVal,PARAM_SHORTCUT,subParamVal);

         if ("corridor" in ANALYSIS_NAME):
            yMax = 12;
            if ("Adaptive" in config.dataDirectoryPath):
               yMax = 3;
         else:
            yMax = 6;
            if ("Adaptive" in config.dataDirectoryPath):
               yMax = 1.8;
         neuronUsageInBins = gaData_module.getNeuronUsageAsBins(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,0,networkTrainingTimeAverage); #networkTrainingTimeAverage
         fileName = "neuronUsage_beforeNovelty_training_t0-{}".format(networkTrainingTimeAverage);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, firstNBins_=50, yMax_=yMax);

         if ("corridor" in ANALYSIS_NAME):
            yMax = 30;
            if ("Adaptive" in config.dataDirectoryPath):
               yMax = 1.6;
         else:
            yMax = 14;
            if ("Adaptive" in config.dataDirectoryPath):
               yMax = 0.3;
         neuronUsageInBins = gaData_module.getNeuronUsageAsBins(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,networkTrainingTimeAverage,noveltyOccurrenceMinTime);
         fileName = "neuronUsage_beforeNovelty_trained_t{}-{}".format(networkTrainingTimeAverage,noveltyOccurrenceMinTime);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, firstNBins_=50, yMax_=yMax);

         if ("corridor" in ANALYSIS_NAME):
            yMax = 60;
            if ("at0.2" in config.dataDirectoryPath):
               yMax = 30;
         else:
            yMax = 16;
            if ("at0.2" in config.dataDirectoryPath):
               yMax = 6;
         neuronUsageInBins = gaData_module.getNeuronUsageAsBins(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,networkTrainingTimeAverage,(int)(networkTrainingTimeAverage + averagePassLength));
         fileName = "neuronUsage_beforeNovelty_trained_firstPass";
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, firstNBins_=50, yMax_=yMax);


         """
         networkTrainingTimeAverage = (int)(math.ceil(crData.getAverageOfAList(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, noveltyOccurrenceMinTime, totalExperimentTime*10))));
         neuronUsageInBins = gaData_module.getNeuronUsageAsBins(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,noveltyOccurrenceMinTime,networkTrainingTimeAverage);
         fileName = "neuronUsage_afterNovelty_training_t{}-{}".format(networkTrainingTimeAverage,noveltyOccurrenceMinTime);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, firstNBins_=50, lastNBins_=10);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName);

         neuronUsageInBins = gaData_module.getNeuronUsageAsBins(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,networkTrainingTimeAverage,totalExperimentTime*10);
         fileName = "neuronUsage_afterNovelty_trained_t{}-{}".format(networkTrainingTimeAverage,totalExperimentTime*10);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, firstNBins_=50, lastNBins_=10);
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName);
         """

#===================== PLOT =============================
if (DO_SINGLE_MEASURE_PLOTS):

   config.outputDirectoryPath = OUTPUT_DIRECTORY + "/"
   config.style.markers = ['cs','bs','rs'];
   config.style.lineStyles = ['-','-','--'];

   legendLabels = [helpers.getHumanReadableParameterString(PARAM_SETS[x]) for x in range(len(PARAM_SETS))];
   xAxisLabel = PARAM_NAME;
   xAxisTicks = X_TICK_VALS;


   #------------------------------------
   # network training time
   #------------------------------------
   yMax = 20;
   if (DO_BLOB_CAMERA_NOIZE):
      yMax = 40;
   visualization.plotNetworkTrainingTimeByPass(config, initialNetworkTrainingTimeData, averagePassLengths, [[0 for v in range(len(PARAM_VALS))] for p in range(len(PARAM_SETS))], legendLabels, xAxisLabel, xAxisTicks, yMax_=yMax);
   visualization.plotNetworkTrainingTimeByPass(config, afterEngChangeNetworkTrainingTimeData, averagePassLengths, noveltyOccurrenceMinTimes, legendLabels, xAxisLabel, xAxisTicks, yMax_=yMax, startTimeLabel_="afterChange"); #y max can be more than 20, because the total time really its more than 40 passes


   #------------------------------------
   # reporting performance
   #------------------------------------

   visualization.plotNoveltyReporting(config, numOfNoveltyTruePositivesData, "true positives", legendLabels, xAxisLabel, xAxisTicks);

   #brokenPlotYMax = 75;
   #brokenPlotYCoord = 1.0;
   brokenPlotYMax = crGraphs.INVALID_VALUE;
   brokenPlotYCoord = crGraphs.INVALID_VALUE;
   yMax = 0.2
   visualization.plotNoveltyReporting(config, numOfNoveltyFalsePositivesData, "false positives", legendLabels, xAxisLabel, xAxisTicks, yMax_=yMax, brokenLinePlot_yMax_=brokenPlotYMax,brokenLinePlot_yCoord=brokenPlotYCoord);


   #------------------------------------
   # network size
   #------------------------------------
   yMax = crGraphs.INVALID_VALUE;
   brokenPlotYMax = crGraphs.INVALID_VALUE;
   brokenPlotYCoord = crGraphs.INVALID_VALUE;

   if (DO_BLOB_CAMERA_NOIZE):
      brokenPlotYMax = 16000;
      brokenPlotYCoord = 1000;
      if ("room" in ANALYSIS_NAME):
         brokenPlotYMax = 9000;
         brokenPlotYCoord = 400;

   visualization.plotNNMetric(config, numOfNeuronsData, "numOfNeurons", constants.TEXT_MAX_NUM_OF_NEURONS, legendLabels, xAxisLabel, xAxisTicks, brokenLinePlot_yMax_=brokenPlotYMax,brokenLinePlot_yCoord=brokenPlotYCoord);

   visualization.plotNNMetric(config, numOfInWeightsData, "numOfInWeights", constants.TEXT_AVERAGE_TOTAL_INPUT_WEIGHTS, legendLabels, xAxisLabel, xAxisTicks);
   visualization.plotNNMetric(config, numOfInWeightsPerNeuronData, "numOfInWeightsPerNeuron", constants.TEXT_AVERAGE_INPUT_WEIGHTS_PER_NEURON, legendLabels, xAxisLabel, xAxisTicks);
   visualization.plotNNMetric(config, numOfConnectionsData, "numOfConnections", constants.TEXT_AVERAGE_TOTAL_CONNECTIONS, legendLabels, xAxisLabel, xAxisTicks);

   visualization.plotNNMetric(config, networkGrowthRateData, "networkGrowthRate", constants.TEXT_GROWTH_RATE, legendLabels, xAxisLabel, xAxisTicks);





